#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

from typing import Union

from pymongo.collection import Collection

_Document = Union[str, int, float, bool, list, dict, None]


def kv_load(c: Collection, key: str) -> _Document:
    record: Union[dict, None] = c.find_one(
        {"_id": key},
        projection={"_id": False, "value": True},
    )
    if record is None:
        return
    return record.get("value")


def kv_save(c: Collection, key: str, value: _Document):
    return c.update_one(
        {"_id": key},
        {"$set": {"value": value}},
        upsert=True,
    )


class KVStore:
    def __init__(self, collection: Collection):
        self._collection = collection

    def load(self, key: str) -> _Document:
        return kv_load(self._collection, key)

    def save(self, key: str, value: _Document):
        return kv_save(self._collection, key, value)



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import argparse

from volkanic.cmdline import CommandOptionDict


def _add_hyphens(parts: list) -> list:
    if not parts:
        return parts
    key = str(parts[0])
    if not key.startswith("-"):
        prefix = "--" if len(key) > 1 else "-"
        key = prefix + key
    return [key] + parts[1:]


class CommandOptionDictMongoish(CommandOptionDict):
    @classmethod
    def _explode(cls, key, val):
        parts = super()._explode(key, val)
        parts = _add_hyphens(parts)
        if len(parts) != 2:
            return parts
        key, val = parts
        if key.startswith("--"):
            return [f"{key}={val}"]


class ArgumentParserMongoish(argparse.ArgumentParser):
    def add_argument_mongoi_host(self):
        self.add_argument("-H", "--host", help="host as in MongoInterface")

    def add_arguments_host_port(self, hostname="127.0.0.1"):
        add = self.add_argument
        add("-H", "--hostname", default=hostname, help="hostname")
        add("-p", "--port", type=int, default=27017, help="port number")

    def add_argument_db_name(self):
        self.add_argument("-d", "--db-name", help="database name")

    def add_argument_coll_name(self):
        self.add_argument("-c", "--coll-name", help="collection name")



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import logging
from typing import Iterable

from pymongo.collection import Collection
from pymongo.errors import OperationFailure

_logger = logging.getLogger(__name__)


def batch_insert(c: Collection, records: Iterable[dict]):
    with c.database.client.start_session() as session:
        with session.start_transaction():
            print("inserting", c, "....")
            # noinspection PyBroadException
            try:
                ir = c.insert_many(records, session=session)
                # commit the transaction: this is implicit if no exception is raised
                session.commit_transaction()
            except OperationFailure:
                _logger.error("transaction failed")
                session.abort_transaction()
                raise
            except Exception:
                session.abort_transaction()
                raise
            return ir


def batch_update(c: Collection, filtr: dict | list, props: dict):
    if isinstance(filtr, list):
        filtr = {"_id": {"$in": filtr}}
    with c.database.client.start_session() as session:
        with session.start_transaction():
            print("deleting", c, filtr, "....")
            # noinspection PyBroadException
            try:
                ur = c.update_many(filtr, {"$set": props})
                session.commit_transaction()
            except OperationFailure:
                _logger.error("transaction failed")
                session.abort_transaction()
                raise
            except Exception:
                session.abort_transaction()
                raise
            return ur


def batch_delete(c: Collection, filtr: dict | list):
    if isinstance(filtr, list):
        filtr = {"_id": {"$in": filtr}}
    with c.database.client.start_session() as session:
        with session.start_transaction():
            print("deleting", c, filtr, "....")
            # noinspection PyBroadException
            try:
                dr = c.delete_many(filtr)
                session.commit_transaction()
            except OperationFailure:
                _logger.error("transaction failed")
                session.abort_transaction()
                raise
            except Exception:
                session.abort_transaction()
                raise
            return dr



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations


import datetime

from pymongo.collection import Collection, ReturnDocument
from pymongo.errors import DuplicateKeyError


class SerialNumber:
    __slots__ = ["prefix", "number", "length"]

    @staticmethod
    def _get_collection() -> Collection:
        raise NotImplementedError

    def __init__(self, prefix="ID", length=6):
        coll = self._get_collection()
        doc = coll.find_one_and_update(
            {"_id": prefix},
            {"$inc": {"i": 1}},
            upsert=True,
            return_document=ReturnDocument.AFTER,
        )
        self.prefix = prefix
        self.number = doc["i"]
        self.length = length

    def __str__(self):
        return self.prefix + str(self.number).zfill(self.length)


class NamedLock(object):
    @staticmethod
    def _get_collection() -> Collection:
        raise NotImplementedError

    def __init__(self, name: str, ttl=12):
        self.name = name
        self.coll = self._get_collection()
        self.ttl = ttl

    def acquire(self):
        now = datetime.datetime.now()
        expire_at = now + datetime.timedelta(seconds=self.ttl)
        self.coll.delete_many({"expire_at": {"$lt": now}})
        record = {"_id": self.name, "expire_at": expire_at}
        try:
            self.coll.insert_one(record)
        except DuplicateKeyError:
            return False
        return True

    def release(self):
        self.coll.delete_one({"_id": self.name})



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import dataclasses
import random
import string
from functools import cached_property
from typing import Union


def _get_random_key(label: str) -> str:
    chars = ["_", label, "_", *random.choices(string.ascii_lowercase, k=20)]
    return "".join(chars)


_Document = Union[str, dict]
_Expression = Union[bool, int, float, str, list, dict]
_ArrayExpression = Union[str, list]


# https://www.mongodb.com/docs/manual/reference/operator/aggregation/not/
# https://www.mongodb.com/docs/manual/reference/operator/aggregation/in/
def not_in(expr: _Expression, array_expr: _ArrayExpression):
    return {"$not": [{"$in": [expr, array_expr]}]}


# https://www.mongodb.com/docs/manual/reference/operator/aggregation/replaceRoot/
def replace_root(*docs: _Document):
    if len(docs) == 1:
        return {
            "$replaceRoot": {
                "newRoot": docs[0],
            }
        }
    return {
        "$replaceRoot": {
            "newRoot": {
                "$mergeObjects": list(docs),
            }
        }
    }


@dataclasses.dataclass
class LookupRecipe:
    """
    Four conceptual stages:
    - lookup {..., _rc: [..., {a: 1, b: 2}]}
    - array flatten {..., _rc: {a: 1, b: 2}}
    - object flatten {..., a: 1, b: 2, _rc: {a: 1, b: 2}}
    - unset {..., a: ..., b: ...}
    """

    from_: str
    local_field: str
    foreign_field: str
    array_idx: int = None  # commonly 0 or -1
    field_map: dict[str, str] = None

    @cached_property
    def _key(self) -> str:
        return _get_random_key(self.from_)

    @property
    def _dollar_key(self) -> str:
        return f"${self._key}"

    # https://www.mongodb.com/docs/manual/reference/operator/aggregation/lookup/
    def get_lookup_stage(self):
        return {
            "$lookup": {
                "from": self.from_,
                "localField": self.local_field,
                "foreignField": self.foreign_field,
                "as": self._key,
            }
        }

    # https://www.mongodb.com/docs/manual/reference/operator/aggregation/unwind/
    def get_array_flatten_stage(self):
        if self.array_idx is None:
            return {
                "$unwind": {
                    "path": self._dollar_key,
                    "preserveNullAndEmptyArrays": True,
                }
            }
        return {
            "$addFields": {
                self._key: {
                    "$ifNull": [
                        {"$arrayElemAt": [self._dollar_key, self.array_idx]},
                        None,
                    ]
                }
            }
        }

    def _infer_object_flatten_stage(self, field_map: dict[str, str] = None):
        if field_map is None:
            return replace_root(self._dollar_key, "$$ROOT")
        mongo_fieldmap = {}
        for new_key, old_key in field_map.items():
            mongo_fieldmap[new_key] = {
                "$ifNull": [f"${self._key}.{old_key}", None],
            }
        return {"$addFields": mongo_fieldmap}

    def get_object_flatten_stage(self):
        return self._infer_object_flatten_stage(self.field_map)

    def get_unset_stage(self):
        return {"$unset": [self._key]}

    # for compatibility; will be removed at ver 0.5.0
    def build_pipeline(self, fieldmap: dict[str, str] = None):
        return [
            self.get_lookup_stage(),
            self.get_array_flatten_stage(),
            self._infer_object_flatten_stage(fieldmap),
            self.get_unset_stage(),
        ]

    def get_stages(self):
        return [
            self.get_lookup_stage(),
            self.get_array_flatten_stage(),
            self.get_object_flatten_stage(),
            self.get_unset_stage(),
        ]


def lookup_and_flatten(
    from_: str,
    local_field: str,
    foreign_field: str,
    array_idx: int = None,
    field_map: dict[str, str] = None,
):
    recipe = LookupRecipe(from_, local_field, foreign_field, array_idx, field_map)
    return recipe.get_stages()


def lookup_unwind_unset(
    from_: str,
    local_field: str,
    foreign_field: str,
    fieldmap: dict[str, str] = None,
    array_idx: int = None,
):
    """for compatibility; will be removed at ver 0.5.0"""
    return lookup_and_flatten(from_, local_field, foreign_field, array_idx, fieldmap)



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

# this module is deprecated
# this module will be removed at ver 0.5.0

from joker.mongodb.query import (
    find_with_renaming,
    find_one_with_renaming,
    find_most_recent,
    find_most_recent_one,
)

_compat_names = [
    find_with_renaming,
    find_one_with_renaming,
    find_most_recent,
    find_most_recent_one,
]



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import logging

from pymongo.monitoring import (
    CommandListener,
    CommandStartedEvent,
    CommandSucceededEvent,
    CommandFailedEvent,
)


class MongoCommandLogger(CommandListener):
    _registered = False
    _logger = logging.getLogger("_mongodb")
    _level = logging.DEBUG

    @staticmethod
    def _fmt_opid(event):
        if event.request_id == event.operation_id:
            return event.request_id
        return "{}.{}".format(event.request_id, event.operation_id)

    @staticmethod
    def _fmt_url(event):
        addr = ":".join(str(s) for s in event.connection_id)
        return "{}/{}".format(addr, event.database_name)

    def started(self, event: CommandStartedEvent):
        if not self._logger.isEnabledFor(self._level):
            return
        parts = [
            self._fmt_opid(event),
            "started",
            self._fmt_url(event),
            event.command,
        ]
        msg = " ".join(str(s) for s in parts)
        self._logger.debug(msg)

    def succeeded(self, event: CommandSucceededEvent):
        if not self._logger.isEnabledFor(self._level):
            return
        parts = [
            self._fmt_opid(event),
            "succeeded",
            int(event.duration_micros / 1000),
        ]
        msg = " ".join(str(s) for s in parts)
        self._logger.debug(msg)

    def failed(self, event: CommandFailedEvent):
        if not self._logger.isEnabledFor(self._level):
            return
        parts = [
            self._fmt_opid(event),
            "failed",
            event.duration_micros,
        ]
        msg = " ".join(str(s) for s in parts)
        self._logger.debug(msg)



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import logging
import subprocess
from pathlib import Path
from urllib.parse import urljoin

import requests

from joker.mongodb.utils import Pathlike

_logger = logging.getLogger(__name__)


def _download(url: str, path: Path) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    with requests.get(url, stream=True) as resp:
        resp.raise_for_status()
        with open(path, "wb") as fout:
            for chunk in resp.iter_content(chunk_size=8192):
                fout.write(chunk)
    _logger.info("downloaded %s", path)


class DumpSuite:
    def __init__(self, dir_: Pathlike, source: str):
        self.dir_ = Path(dir_)
        self.source = source

    def fetch(self, db_name: str, collection_name: str):
        rel_paths = [
            f"{db_name}/{collection_name}.bson.gz",
            f"{db_name}/{collection_name}.metadata.json.gz",
        ]
        for path in rel_paths:
            url = urljoin(self.source, path)
            _download(url, self.dir_ / path)

    def batch_fetch(self, names: list[str]):
        for name in names:
            db_name, collection_name = name.split(".", maxsplit=1)
            self.fetch(db_name, collection_name)

    def restore(self, host: str = "mongodb", port: int = 27017):
        cmd = [
            "mongorestore",
            f"--dir={self.dir_}",
            "--gzip",
            "--drop",
            f"--host={host}",
            f"--port={port}",
        ]
        subprocess.run(cmd, check=True)



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import logging
from typing import Iterable

from pymongo.collection import Collection
from pymongo.cursor import Cursor

_logger = logging.getLogger(__name__)


def _namemap_to_project(namemap: dict):
    project = {}
    for new_name, old_name in namemap.items():
        if not old_name.startswith("$"):
            old_name = "$" + old_name
        project[new_name] = {"$ifNull": [old_name, None]}
    return project


def _namemap_from_fieldlist(fieldlist: list):
    return {k.split(".")[-1]: k for k in fieldlist}


def find_with_renaming(coll: Collection, filtr: dict, namemap: dict, sort: dict = None):
    pipelines = [
        {"$match": filtr},
        {"$sort": sort or {"_id": -1}},
        {"$project": _namemap_to_project(namemap)},
    ]
    return coll.aggregate(pipelines)


def find_one_with_renaming(
    coll: Collection, filtr: dict, namemap: dict, sort: dict = None
):
    pipelines = [
        {"$match": filtr},
        {"$sort": sort or {"_id": -1}},
        {"$limit": 1},
        {"$project": _namemap_to_project(namemap)},
    ]
    _logger.debug("pipelines: %s", pipelines)
    docs = list(coll.aggregate(pipelines))
    if docs:
        return docs[0]


def find_most_recent(c: Collection, **kwargs) -> Cursor:
    kwargs.setdefault("limit", 10)
    kwargs.setdefault("sort", [("$natural", -1)])
    return c.find(**kwargs)


def find_most_recent_one(c: Collection, **kwargs) -> dict:
    kwargs.setdefault("sort", [("$natural", -1)])
    return c.find_one(**kwargs)


def find_unique(c: Collection, keys: list[str]) -> Iterable[dict]:
    pipeline = [{"$group": {"_id": {k: f"${k}" for k in keys}}}]
    for record in c.aggregate(pipeline):
        yield record["_id"]


def find_unique_tuples(c: Collection, keys: list[str]) -> Iterable[tuple]:
    cursor = find_unique(c, keys)
    for record in cursor:
        yield tuple(record[k] for k in keys)



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

from typing import Iterable

from joker.mongodb.tools.pagination import RawResultDict, PaginatedResult

# this module has been renamed to pagination.py
# this module will be removed at ver 0.5.0


def compact_paginated_result(cursor: Iterable[RawResultDict]) -> PaginatedResult:
    # this function is deprecated
    # this function will be removed at ver 0.5.0
    return PaginatedResult.from_raw(cursor)



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import datetime

from bson import ObjectId


def oid_filter_by_datetime(
    start: datetime.datetime = None, end: datetime.datetime = None
) -> dict:
    # this function is deprecated
    # this function will be removed at ver 0.5.0
    filtr = {}
    if start is not None:
        filtr["$gt"] = ObjectId.from_datetime(start)
    if end:
        filtr["$lt"] = ObjectId.from_datetime(end)
    return filtr


def oid_filter_recent(days=30, seconds=0):
    # this function is deprecated
    # this function will be removed at ver 0.5.0
    delta = datetime.timedelta(days=days, seconds=seconds)
    start = datetime.datetime.now() - delta
    return oid_filter_by_datetime(start, None)



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import logging
import threading
import time
import traceback
from collections import UserDict
from collections import defaultdict
from typing import Callable

import pymongo
from bson import Timestamp, json_util, ObjectId
from pymongo import MongoClient
from pymongo.database import Database

_logger = logging.getLogger(__name__)


class OplogRecord(UserDict):
    @property
    def op(self) -> str:
        return self.data.get("op")

    @property
    def ns(self) -> str:
        return self.data.get("ns")

    @property
    def ts(self) -> Timestamp:
        return self.data.get("ts")

    def to_mongo_doc(self):
        doc = self.data.copy()
        doc["ns"] = self.ns
        if "o" in doc:
            doc["o"] = json_util.dumps(doc["o"])
        return doc

    @classmethod
    def from_mongo_doc(cls, doc: dict):
        doc = dict(doc)
        if "o" in doc:
            doc["o"] = json_util.loads(doc["o"])
        return cls(doc)

    @property
    def upstream_id(self) -> str:
        try:
            return str(self.data["o"]["_id"])
        except (KeyError, TypeError):
            pass
        try:
            return str(self.data["o2"]["_id"])
        except (KeyError, TypeError):
            pass

    @property
    def id_(self) -> ObjectId:
        if id_ := self.data.get("_id"):
            return ObjectId(id_)


# Caution: do NOT run multiple threads / processes of this!!
# TODO: consider thread safety
class OplogTailer(object):
    _ns_exclude = {}
    _db_exclude = {"config", "local", "admin"}
    record_cls = OplogRecord

    def __init__(
        self,
        upstream_client: MongoClient,
        ts: Timestamp,
        ns_pattern: str = None,
        ns_exclude: str = None,
    ):
        """
        Args:
            upstream_client:
            ts: starting timestamp (seconds since epoch), e.g. 1642477123
            ns_pattern: regex
            ns_exclude: regex
        """
        db = upstream_client.get_database("local")
        self.oplog_coll = db.get_collection("oplog.rs")
        self.ts = ts
        self.ns_pattern = ns_pattern
        self.ns_exclude = ns_exclude
        self._cursor = self._get_cursor()

    def _get_where_clause(self):
        tests = ['this.op != "n"']
        if self.ns_pattern is not None:
            tests.append("this.ns.match(/{}/)".format(self.ns_pattern))
            # tests.append(f'this.ns.match(/{self.ns_pattern}/)')
        if self.ns_exclude is not None:
            tests.append("!this.ns.match(/{}/)".format(self.ns_exclude))
            # tests.append(f'!this.ns.match(/{self.ns_exclude}/)')
        return "&&".join(tests)

    def _get_cursor(self):
        filtr = {"ts": {"$gt": self.ts}}
        # https://pymongo.readthedocs.io/en/stable/examples/tailable.html
        # TODO: consider using $expr for mongo >= 3.6
        _logger.info("OplogTailer._get_cursor, filtr=%s", filtr)
        cursor = self.oplog_coll.find(
            filtr,
            oplog_replay=True,
            cursor_type=pymongo.CursorType.TAILABLE_AWAIT,
        )
        where_clause = self._get_where_clause()
        if where_clause:
            cursor = cursor.where(where_clause)
        return cursor

    def _reset_cursor(self):
        self._cursor.close()
        self._cursor = self._get_cursor()

    def _check_ns(self, doc: dict):
        ns = doc.get("ns")
        if not ns:
            return False
        if ns in self._ns_exclude:
            return False
        db_name = ns.split(".")[0]
        if db_name in self._db_exclude:
            return False
        return True

    def _fetch_next(self):
        try:
            return self._cursor.next()
        except (StopIteration, ConnectionResetError):
            self._reset_cursor()

    def __iter__(self):
        return self

    def __next__(self) -> "record_cls":
        while True:
            doc = self._fetch_next()
            if doc is None:
                time.sleep(1)
                continue
            if not self._check_ns(doc):
                continue
            self.ts = doc.get("ts")
            return self.record_cls(doc)


class ChangeStreamRegistry:
    def __init__(self, db: Database):
        self.db = db
        self.handlers = defaultdict(list)

    @staticmethod
    def apply(handler: Callable, coll_name: str, event: dict):
        try:
            id_ = event["documentKey"]["_id"]
        except KeyError:
            id_ = None
        _logger.info(
            "applying %s() to change event of collection %r, %s",
            getattr(handler, "__name__", None) or str(handler),
            coll_name,
            id_,
        )
        # noinspection PyBroadException
        try:
            handler(event)
        except Exception:
            traceback.print_exc()

    def watch(self, coll_name: str):
        coll = self.db[coll_name]
        cursor = coll.watch()
        for event in cursor:
            for handler in self.handlers.get(coll_name):
                self.apply(handler, coll_name, event)

    def register(self, coll_name: str, handler: Callable):
        self.handlers[coll_name].append(handler)

    def execute(self):
        threads = []
        for coll_name, handlers in self.handlers.items():
            args = tuple([coll_name])
            thr = threading.Thread(target=self.watch, args=args)
            thr.start()
            threads.append(thr)
        for thr in threads:
            thr.join()



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import dataclasses
from typing import TypedDict, Iterable


class _CountDict(TypedDict):
    total: int


class RawResultDict(TypedDict):
    documents: list[dict]
    counts: list[_CountDict]


@dataclasses.dataclass
class QueryParams:
    skip: int = 0
    limit: int = 10
    sort: str = "_id"
    order: int = -1
    keyword: str = None

    def get_facet_stage(self):
        return {
            "$facet": {
                "documents": [
                    {"$skip": self.skip},
                    {"$limit": self.limit},
                ],
                # total count of documents
                "counts": [{"$count": "total"}],
            }
        }

    def get_sort_stage(self):
        return {"$sort": {self.sort: self.order}}


    def get_pagination_pipeline(self):
        return [
            self.get_sort_stage(),
            self.get_facet_stage(),
        ]


@dataclasses.dataclass
class PaginatedResult:
    items: list[dict]
    total: int

    @classmethod
    def from_raw(
        cls: type[PaginatedResult], raw: Iterable[RawResultDict]
    ) -> PaginatedResult:
        # TODO: use typing.Self -- requires python >= 3.11
        # from typing import Self
        # from_raw(cls: type[Self], raw: Iterable[RawResultDict]) -> Self:
        result: RawResultDict = list(raw)[0]
        counts = result["counts"]
        total = counts[0]["total"] if counts else 0
        return PaginatedResult(items=result["documents"], total=total)

    def to_dict(self) -> dict:
        return dataclasses.asdict(self)



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import csv
import datetime
import functools
import logging
import sys
import traceback
from collections import defaultdict
from typing import Union

import pymongo.errors
from bson import ObjectId
from pymongo.collection import Collection
from pymongo.cursor import Cursor
from pymongo.database import Database
from volkanic.utils import printerr

from joker.mongodb import utils

_logger = logging.getLogger(__name__)


# too smart ..., deprecate?
def projection_(*keys, value=True, smart=True):
    if smart and len(keys) == 1:
        keys = keys[0].split()
    return {k: bool(value) for k in keys}


def _show_call_stack(func):
    @functools.wraps(func)
    def ret_func(*args, **kwargs):
        print(func, file=sys.stderr)
        rv = func(*args, **kwargs)
        traceback.print_stack()
        return rv

    return ret_func


def patch_mongo_for_debug(wrapper_func=None):
    if wrapper_func is None:
        wrapper_func = _show_call_stack
    Collection.find = wrapper_func(Collection.find)
    Collection.find_one = wrapper_func(Collection.find_one)
    Collection.update = wrapper_func(Collection.update)  # noqa
    Collection.update_one = wrapper_func(Collection.update_one)
    Collection.update_many = wrapper_func(Collection.update_many)


def gridfs_copy(source_fs, target_fs, _id):
    if isinstance(_id, str):
        _id = ObjectId(_id)
    print('copying gridfs file:', _id)
    source_file = source_fs.find_one({'_id': _id})
    if not source_file:
        raise ValueError('not found: _id = {}'.format(_id))
    target_file = target_fs.new_file(_id=_id)
    target_file.write(source_file)
    target_file.close()


def export_records_to_csv(records, outpath):
    fields = set()
    for rec in records:
        fields.update(rec)
    header = list(fields)
    header.sort()
    with open(outpath, 'w') as fout:
        writer = csv.writer(fout)
        writer.writerow(header)
        for rec in records:
            row = [rec.get(k, '') for k in header]
            writer.writerow(row)


def _insert(target_coll: Collection, record: dict, identifier=None):
    identifier = identifier or str(record.get('_id'))
    try:
        target_coll.insert_one(record)
    except pymongo.errors.DuplicateKeyError:
        printerr('Dup:', target_coll, identifier)
    else:
        printerr('OK:', target_coll, identifier)
    return record


def copy_one(source_coll, target_coll, filtr: dict, updates: dict = None, **kwargs):
    record = source_coll.find_one(filtr, **kwargs)
    if not record:
        printerr('NotFound:', source_coll, filtr)
        return
    if updates:
        record.update(updates)
    return _insert(target_coll, record, identifier=filtr)


def copy_many(source_coll, target_coll, filtr: dict, updates: dict = None, **kwargs):
    for record in source_coll.find(filtr, **kwargs):
        if updates:
            record.update(updates)
        _insert(target_coll, record)


def find_field_names(coll: Collection, retry: int = 10):
    keys = set()
    name_count = 0
    for i in range(50):
        doc = coll.find_one(skip=i, sort=[('_id', -1)])
        keys.update(doc)
        if len(keys) == name_count:
            retry -= 1
        else:
            name_count = len(keys)
    return keys


class DatabaseWrapper:
    def __init__(self, db: Database):
        self.db = db

    def inspect_storage_sizes(self):
        return utils.inspect_mongo_storage_sizes(self.db)

    def print_storage_sizes(self):
        return utils.print_mongo_storage_sizes(self.db)


class CollectionWrapper:
    def __init__(self, coll: Collection, filtr=None, projection=None):
        self.coll = coll
        self.filtr = filtr or {}
        self.projection = projection

    def exist(self, filtr: Union[ObjectId, dict]):
        return self.coll.find_one(filtr, projection=[])

    def find_by_oid_range(
            self, start: Union[ObjectId, None],
            end: Union[ObjectId, None]) -> Cursor:
        _id_conf = {}
        if start:
            _id_conf['$gt'] = start
        if end:
            _id_conf['$lt'] = end
        filtr = self.filtr.copy()
        filtr['_id'] = _id_conf
        return self.coll.find(filtr, projection=self.projection)

    def find_by_oid_time_range(
            self, start: Union[datetime.datetime, None],
            end: Union[datetime.datetime, None]) -> Cursor:
        _id_conf = {}
        if start:
            _id_conf['$gt'] = ObjectId.from_datetime(start)
        if end:
            _id_conf['$lt'] = ObjectId.from_datetime(end)
        filtr = self.filtr.copy()
        if _id_conf:
            filtr['_id'] = _id_conf
        _logger.debug(filtr)
        return self.coll.find(filtr, projection=self.projection)

    def find_recent_by_time(self, days=30, seconds=0) -> Cursor:
        delta = datetime.timedelta(days=days, seconds=seconds)
        start = datetime.datetime.now() - delta
        return self.find_by_oid_time_range(start, None)

    def find_recent_by_count(self, count=50) -> Cursor:
        cursor = self.coll.find(self.filtr, projection=self.projection)
        return cursor.sort([('_id', -1)]).limit(count)

    def find_most_recent_one(self) -> dict:
        recs = list(self.find_recent_by_count(1))
        if recs:
            return recs[0]

    def _replace(self, records, uk='_id'):
        """
        :param records: a list of dicts
        :param uk: unique key
        """
        for rec in records:
            uk_val = rec.get(uk)
            self.coll.replace_one({uk: uk_val}, rec)

    def _update(self, records, uk='_id'):
        """
        :param records: a list of dicts
        :param uk: unique key
        """
        for rec in records:
            uk_val = rec.get(uk)
            self.coll.update_one({uk: uk_val}, {'$set': rec})

    def _insert(self, records):
        if records:
            self.coll.insert_many(records, ordered=False)

    @staticmethod
    def _check_for_uniqueness(records, uk):
        vals = [r.get(uk) for r in records]
        uniq_vals = set(vals)
        if len(vals) != len(uniq_vals):
            raise ValueError('records contain duplicating keys')

    def upsert_with_index(self, records, uk='_id', replace=False):
        """Batch insert or update
        requires single-field unique index on `uk`
        Similar to `INESRT ... ON DUPLICATE KEY UPDATE ...` in MySQL
        Caution:
        - repeats in records lead to unpredictable results
        - not atomic
        :param records: a list of dicts
        :type records: list
        :param uk: unique key
        :type uk: str
        :param replace: use coll.replace instead of collection.update
        :type replace: bool
        """
        if not records:
            return
        dup_records = []
        try:
            self._insert(records)
        except pymongo.errors.BulkWriteError as exc:
            other_errors = []
            for err in exc.details['writeErrors']:
                if err['code'] == 11000:
                    dup_records.append(err['op'])
                else:
                    other_errors.append(err)
            if other_errors or exc.details['writeConcernErrors']:
                exc.details['writeErrors'] = other_errors
                raise exc
        if replace:
            self._replace(records, uk=uk)
        else:
            self._update(records, uk=uk)

    def upsert_without_index(self, records, uk='_id', replace=False):
        """Batch insert or update
        single-field unique index on `uk` is not required
        Caution:
        - repeats in records lead to unpredictable results
        - not atomic
        :param records: a list of dicts
        :type records: list
        :param uk: unique key
        :type uk: str
        :param replace: use coll.replace instead of collection.update
        :type replace: bool
        """
        if not records:
            return
        self._check_for_uniqueness(records, uk)
        insert_records = []
        update_records = []
        existing_vals = {r.get(uk) for r in self.coll.find({}, ['uk'])}
        for rec in records:
            uk_val = rec.get(uk)
            if uk_val in existing_vals:
                update_records.append(rec)
            else:
                insert_records.append(rec)
        self._insert(insert_records)
        if replace:
            self._replace(update_records, uk=uk)
        else:
            self._update(update_records, uk=uk)

    def make_fusion_record(self):
        fusion_record = {}
        contiguous_stale_count = -1
        for skip in range(1000):
            record = self.coll.find_one(sort=[('$natural', -1)], skip=skip)
            if not record:
                continue
            contiguous_stale_count += 1
            for key, val in record.items():
                if not record.get(key):
                    fusion_record[key] = val
                    contiguous_stale_count = -1
            if contiguous_stale_count > 10:
                return fusion_record
        return fusion_record

    def query_uniq_values(self, fields: list, limit=1000):
        latest = [('_id', -1)]
        records = self.coll.find(sort=latest, projection=fields, limit=limit)
        uniq = defaultdict(set)
        for key in fields:
            for rec in records:
                val = rec.get(key)
                uniq[key].add(val)
        return uniq

    def batch_update_a_field_with_log(self, field_name, old_value, new_value, log=None):
        upd = {'$set': {field_name: new_value}}
        if log:
            upd['$push'] = {'log': log}
        ur = self.coll.update_many({field_name: old_value}, upd)
        s = f'{field_name}={old_value}'
        print(s, ur.matched_count, ur.modified_count)



#!/usr/bin/env python3
# coding: utf-8
import os
import re

import setuptools
from setuptools import find_namespace_packages

# CAUTION:
# Do NOT import your package from your setup.py

_nsp = "joker"
_pkg = "mongodb"
_names = [_nsp, _pkg]
_names = [s for s in _names if s]


def read(filename):
    with open(filename) as f:
        return f.read()


def _find_version():
    names = _names + ["__init__.py"]
    path = os.path.join(*names)
    root = os.path.dirname(__file__)
    path = os.path.join(root, path)
    regex = re.compile(r"""^__version__\s*=\s*('|"|'{3}|"{3})([.\w]+)\1\s*(#|$)""")
    with open(path) as fin:
        for line in fin:
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            mat = regex.match(line)
            if mat:
                return mat.groups()[1]
    raise ValueError("__version__ definition not found")


config = {
    "name": "joker-mongodb",
    "version": _find_version(),
    "description": "access mongodb with handy utilities and fun",
    "keywords": "",
    "url": "https://github.com/frozflame/joker-mongodb",
    "author": "frozflame",
    "author_email": "frozflame@outlook.com",
    "license": "GNU General Public License (GPL)",
    "packages": find_namespace_packages(include=["joker.*"]),
    "zip_safe": False,
    "install_requires": read("requirements.txt"),
    "python_requires": ">=3.7.0",
    "classifiers": [
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
    ],
    # ensure copy static file to runtime directory
    "include_package_data": True,
    "long_description": read("README.md"),
    "long_description_content_type": "text/markdown",
}

if _nsp:
    config["namespace_packages"] = [_nsp]

setuptools.setup(**config)



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

from gridfs import GridFS


def _get_gridfs_namespace(fs: GridFS):
    coll = getattr(fs, "_GridFS__files")
    db = coll.database
    return f"{db.name}.{coll.name}"



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import datetime

import pymongo.errors
from pymongo import ReturnDocument
from pymongo.collection import Collection


class TransactionHelper:
    # this class is deprecated
    # this class will be removed at ver 0.5.0

    def __init__(self, coll: Collection):
        self.coll = coll

    def lock(self, name: str, ttl: int = 60):
        now = datetime.datetime.now()
        expire_at = now + datetime.timedelta(seconds=ttl)
        self.coll.delete_many(
            {"expire_at": {"$lt": now}, "type": "lock"},
        )
        record = {
            "_id": name,
            "type": "lock",
            "expire_at": expire_at,
        }
        try:
            self.coll.insert_one(record)
        except pymongo.errors.DuplicateKeyError:
            return False
        return True

    def unlock(self, name: str):
        self.coll.delete_one({"_id": name, "type": "lock"})

    def get_next_serial_code(self, prefix="ID", length=6):
        doc = self.coll.find_one_and_update(
            {"_id": prefix, "type": "serial"},
            {"$inc": {"i": 1}},
            upsert=True,
            return_document=ReturnDocument.AFTER,
        )
        return prefix + str(doc["i"]).zfill(length)



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import datetime

from bson import ObjectId


def in_(*values):
    return {"$in": list(values)}


def exclude(*keys: str):
    return dict.fromkeys(keys, False)


def _dict_from_keys(keys: tuple[str, ...], value):
    if not keys:
        return value
    return dict.fromkeys(keys, value)


_py_false_vals = [0, "", 0.0, [], {}, False, None]
_js_false_vals = [0, "", 0.0, False, None]


def py_true(*keys: str) -> dict:
    return _dict_from_keys(keys, {"$nin": _py_false_vals})


def py_false(*keys: str) -> dict:
    return _dict_from_keys(keys, {"$in": _py_false_vals})


def js_true(*keys: str) -> dict:
    return _dict_from_keys(keys, {"$nin": _js_false_vals})


def js_false(*keys: str) -> dict:
    return _dict_from_keys(keys, {"$in": _js_false_vals})


def exists(*keys: str) -> dict:
    return _dict_from_keys(keys, {"$exists": True})


def not_exists(*keys: str) -> dict:
    return _dict_from_keys(keys, {"$exists": False})


def recent(days=30, seconds=0):
    delta = datetime.timedelta(days=days, seconds=seconds)
    return datetime.datetime.now() - delta


def oid_range(gte: datetime.datetime = None, lte: datetime.datetime = None):
    filtr = {"$type": "objectId"}
    if gte is not None:
        filtr["$gte"] = ObjectId.from_datetime(gte)
    if lte:
        filtr["$lte"] = ObjectId.from_datetime(lte)
    return filtr



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import os.path
from os import PathLike
from typing import Union

import bson.json_util
import volkanic.utils
from joker.cast.numeric import human_filesize
from joker.textmanip.tabular import tabular_format
from pymongo import MongoClient
from pymongo.database import Database

from joker.mongodb.logger import MongoCommandLogger

_compat = [
    MongoCommandLogger,
]

Pathlike = Union[str, PathLike]


def is_in_replset(mongo: MongoClient) -> bool:
    db = mongo.get_database("admin")
    return "replication" in db.command("getCmdLineOpts")["parsed"]


def inspect_mongo_storage_sizes(target: Union[MongoClient, Database]):
    if isinstance(target, MongoClient):
        return {r["name"]: r["sizeOnDisk"] for r in target.list_databases()}
    size_of_collections = {}
    for coll_name in target.list_collection_names():
        info = target.command("collStats", coll_name)
        size_of_collections[info["ns"]] = info["storageSize"]
    return size_of_collections


def print_mongo_storage_sizes(target: Union[MongoClient, Database]):
    s_rows = list(inspect_mongo_storage_sizes(target).items())
    s_rows.sort(key=lambda r: r[1], reverse=True)
    rows = []
    for k, v in s_rows:
        num, unit = human_filesize(v)
        rows.append([round(num), unit, k])
    for row in tabular_format(rows):
        print(*row)


def indented_json_dumps(obj, **kwargs):
    kwargs.setdefault("dumps", bson.json_util.dumps)
    return volkanic.utils.indented_json_dumps(obj, **kwargs)


def indented_json_print(obj, **kwargs):
    kwargs.setdefault("dumps", indented_json_dumps)
    return volkanic.utils.indented_json_print(obj, **kwargs)


def infer_coll_triple_from_filename(path: str):
    """
    >>> p = "somedir/local.retail.customers.6789.json"
    >>> infer_coll_triple_from_filename(p)
    ['local', 'retail', 'customers']
    """
    filename = os.path.split(path)[1]
    coll_fullname = filename.rsplit(".", 2)[0]
    return coll_fullname.split(".", 2)


def infer_params(mongo: MongoClient):
    params = dict(zip(["host", "port"], mongo.address))
    params.update(mongo._MongoClient__options._options)  # noqa
    return params
